BEGIN_GTBMOUSENS


inline bool Box3::is_order_correct() const
{
	return((_min_pt.x() <= _max_pt.x()) &&
	       (_min_pt.y() <= _max_pt.y()) &&
	       (_min_pt.z() <= _max_pt.z()));
}


inline Box3::Box3()
{
}


inline Box3::Box3(const Box3 &b)
	: _min_pt(b._min_pt),
	  _max_pt(b._max_pt)
{
	assert(is_order_correct());
}


inline Box3::Box3(const Point3 &min_pt, const Point3 &max_pt)
	: _min_pt(min_pt),
	  _max_pt(max_pt)
{
	assert(is_order_correct());
}


inline Box3::Box3(real_t xmin, real_t ymin, real_t zmin,
	   real_t xmax, real_t ymax, real_t zmax)
	: _min_pt(xmin, ymin, zmin),
	  _max_pt(xmax, ymax, zmax)
{
	assert(is_order_correct());
}


inline Box3::Box3(const Point3 &c, real_t lx, real_t ly, real_t lz)
	: _min_pt(c.x() - lx / 2.0, c.y() - ly / 2.0, c.z() - lz / 2.0),
	  _max_pt(c.x() + lx / 2.0, c.y() + ly / 2.0, c.z() + lz / 2.0)
{
	assert(is_order_correct());
}


inline Box3 &Box3::operator=(const Box3 &b)
{
	if (&b != this) {
		_min_pt = b._min_pt;
		_max_pt = b._max_pt;
		assert(is_order_correct());
	}
	return *this;
}


inline bool Box3::operator==(const Box3 &b) const
{
	return ((_min_pt == b._min_pt) &&
		(_max_pt == b._max_pt));
}


inline bool Box3::operator!=(const Box3 &b) const
{
	return !(*this == b);
}


inline bool Box3::is_empty() const
{
	return ((_min_pt.x() >= _max_pt.x()) ||
		(_min_pt.y() >= _max_pt.y()) ||
		(_min_pt.z() >= _max_pt.z()));
}


inline const Point3 &Box3::min_point() const
{
	return _min_pt;
}


inline const Point3 &Box3::max_point() const
{
	return _max_pt;
}


inline real_t Box3::x_min() const
{
	return _min_pt.x();
}


inline real_t Box3::y_min() const
{
	return _min_pt.y();
}


inline real_t Box3::z_min() const
{
	return _min_pt.z();
}


inline real_t Box3::x_max() const
{
	return _max_pt.x();
}


inline real_t Box3::y_max() const
{
	return _max_pt.y();
}


inline real_t Box3::z_max() const
{
	return _max_pt.z();
}


inline real_t Box3::x_length() const
{
	return _max_pt.x() - _min_pt.x();
}


inline real_t Box3::y_length() const
{
	return _max_pt.y() - _min_pt.y();
}


inline real_t Box3::z_length() const
{
	return _max_pt.z() - _min_pt.z();
}


inline real_t Box3::diagonal_length() const
{
	return (_max_pt - _min_pt).length();
}


inline real_t Box3::volume() const
{
	return x_length() * y_length() * z_length();
}


inline Point3 Box3::centroid() const
{
	return Point3((_min_pt.x() + _max_pt.x()) / 2.0,
		      (_min_pt.y() + _max_pt.y()) / 2.0,
		      (_min_pt.z() + _max_pt.z()) / 2.0);
}


inline bool Box3::contains(const Point3 &p) const
{
	return ((p.x() >= _min_pt.x()) && (p.x() <= _max_pt.x()) &&
		(p.y() >= _min_pt.y()) && (p.y() <= _max_pt.y()) &&
		(p.z() >= _min_pt.z()) && (p.z() <= _max_pt.z()));
}


inline void Box3::reset(const Point3 &min_pt, const Point3 &max_pt)
{
	_min_pt = min_pt;
	_max_pt = max_pt;
	assert(is_order_correct());
}


inline void Box3::reset(real_t xmin, real_t ymin, real_t zmin,
		 real_t xmax, real_t ymax, real_t zmax)
{
	_min_pt.reset(xmin, ymin, zmin);
	_max_pt.reset(xmax, ymax, zmax);
	assert(is_order_correct());
}

inline void Box3::set_min(int axis, double value)
{
	_min_pt[axis] = value;
}

inline void Box3::set_max(int axis, double value)
{
	_max_pt[axis] = value;
}

inline void Box3::scale(real_t s)
{
	Point3 origin = centroid();
	_min_pt.scale(origin, s);
	_max_pt.scale(origin, s);
	assert(is_order_correct());
}


inline Point3 Box3::vertex(unsigned i) const
{
	assert(i < 8);
	real_t x = (i & RIGHT_MASK) ? _max_pt.x() : _min_pt.x();
	real_t y = (i & UP_MASK) ?    _max_pt.y() : _min_pt.y();
	real_t z = (i & FRONT_MASK) ? _max_pt.z() : _min_pt.z();
	return Point3(x, y, z);
}


inline Point3 Box3::vertex(unsigned f, unsigned i) const
{
	assert(f < 6);
	assert(i < 4);
	return vertex(_vertex_indices[f][i]);
}


inline Vector3 Box3::normal(unsigned i) const
{
	assert(i < 6);
	return Point3::normal(vertex(_vertex_indices[i][0]),
			      vertex(_vertex_indices[i][1]),
			      vertex(_vertex_indices[i][2]));
}


inline void Box3::read(FILE *fp)
{
	_min_pt.read(fp);
	_max_pt.read(fp);
}


inline void Box3::write(FILE *fp) const
{
	_min_pt.write(fp);
	_max_pt.write(fp);
}


inline double Box3::distance(const Point3 &p) const
{
	Vector3 dl = _min_pt - p;
	Vector3 dr = p - _max_pt;
	for (int i = 0; i < 3; i++) {
		if (dl[i] < 0) {
			dl[i] = 0;
		}
		if (dr[i] < 0) {
			dr[i] = 0;
		}
	}
	Vector3 diff = dl + dr;
	double d = diff.length();
	return d;
}


inline Box3 Box3::bounding_box(const Point3 &a,
			       const Point3 &b,
			       const Point3 &c)
{
	return Box3(min(min(a[0], b[0]), c[0]),
		    min(min(a[1], b[1]), c[1]),
		    min(min(a[2], b[2]), c[2]),
		    max(max(a[0], b[0]), c[0]),
		    max(max(a[1], b[1]), c[1]),
		    max(max(a[2], b[2]), c[2]));
}


END_GTBMOUSENS
