#ifndef GTB_CAMERA_INCLUDED
#define GTB_CAMERA_INCLUDED

//#include "gtb/common.hpp"
//#include "gtb/math/math.hpp"
#include "gtbmousecommon.h"
#include "point3.hpp"
#include "vector3.hpp"
#include "coordinate_system.hpp"
#include "box3.hpp"
#include "plane.hpp"
//#include "polygon3.hpp"
#include "viewport.hpp"
#include "light.hpp"


BEGIN_GTBMOUSENS


class Camera {
public:
	enum Type { PERSPECTIVE, ORTHOGRAPHIC };

	Camera();

	Camera(const Point3 &origin,
	       const Vector3 &towards,
	       const Vector3 &up,
	       real_t x_fov,
	       real_t y_fov,
	       real_t near_distance,
	       real_t far_distance);

	Camera(const Camera &cam);

	Camera(const Camera &cam,
	       unsigned i,
	       unsigned j,
	       unsigned n_rows,
	       unsigned n_cols);

	Camera &operator=(const Camera &cam);

	void reset(const Point3 &origin,
		   const Vector3 &towards,
		   const Vector3 &up);

	void reset(const Point3 &origin,
		   const Vector3 &towards,
		   const Vector3 &up,
		   real_t x_fov,
		   real_t y_fov,
		   real_t near_distance,
		   real_t far_distance);

        void reset_fov(real_t fovx, real_t fovy);

	const Point3 &origin() const;
	Vector3 towards() const;
	const Vector3 &backwards() const;
	const Vector3 &up() const;
	Vector3 down() const;
	const Vector3 &right() const;
	Vector3 left() const;
	real_t x_fov() const;
	real_t y_fov() const;
	real_t aspect() const;
	real_t near_distance() const;
	real_t far_distance() const;
	const CoordinateSystem &coordinate_system() const;

	void load(const Viewport &viewport) const;
	void load(const Viewport &viewport,
		  const std::vector<Light> &headlights) const;
	//void render() const;
	void render(real_t scale) const;

	void view_all(const Box3 &box);
	void init_exterior_view(const Box3 &box);
	void init_interior_view(const Box3 &box);
	void init_image_view();

	void rotate(const Line3 &l, real_t theta);
	void pitch(real_t angle);
	void yaw(real_t angle);
	void roll(real_t angle);

	void translate(const Vector3 &t);
	void move_to(const Point3 &p);

	bool sees(const Point3 &point) const;
	bool sees(const Box3 &box) const;
	bool sees_ignoring_near_plane(const Box3 &b) const;
	bool contains(const Box3 &box) const;

	bool read(FILE *fp);
	void write(FILE *fp) const;

	bool read(const char *file_name);
	void write(const char *file_name) const;

	Matrix4 matrix() const;
	Matrix4 inverse_matrix() const;

	Point3 world_to_camera(const Point3 &p) const;
	Point3 camera_to_world(const Point3 &p) const;

  	friend std::ostream &operator<<(std::ostream &s, const Camera &c);

protected:

	void setup_viewport(const Viewport &viewport) const;
	void setup_frustum() const;
	void load_headlights(const std::vector<Light> &headlights) const;
	void setup_look_at() const;

	CoordinateSystem _cs;
	real_t _x_fov, _y_fov;
	real_t _near_distance, _far_distance;
	Plane _planes[6];
	Point3 _corners[8];
	unsigned _row, _col;
	unsigned _n_rows, _n_cols;
	Type _type;
};


END_GTBMOUSENS

#ifndef OUTLINE
#include "camera.ipp"
#endif

#endif // GTB_CAMERA_INCLUDED
