#include "gtbmouse.h"
#ifndef WIN32
#include <gtb/graphics/view.hpp>
#include <gtb/graphics/line3.hpp>
#include <gtb/math/math.hpp>
#endif // WIN32


#ifdef OUTLINE
#define inline
#include <gtb/graphics/view.ipp>
#undef inline
#endif


using namespace std;


BEGIN_GTBMOUSENS


View::View()
{
}


View::View(const Camera &cam,
	   const Viewport &vp)
	: _camera(cam),
	  _viewport(vp)
{
}


View::View(const View &view)
	: _camera(view._camera),
	  _viewport(view._viewport)
{
}


void View::rotate_world(const Point3 &origin, int dx, int dy)
{
	if ((dx == 0) && (dy == 0)) {
		return;
	}
	// find angle
	real_t vx = (real_t) dx / (real_t) _viewport.width();
	real_t vy = (real_t) -dy / (real_t) _viewport.height();
	real_t theta = 2.0 * M_PI * (fabs(vx) + fabs(vy));

	// find axis
	Vector3 v = vx * _camera.right() + vy * _camera.up();
	Vector3 axis = _camera.towards().cross(v);
	Line3 l(origin, axis);

	_camera.rotate(l, theta);
}


void View::translate_world(const Point3 &origin, int dx, int dy)
{
	if ((dx == 0) && (dy == 0)) {
		return;
	}
	real_t l = Point3::distance(origin, _camera.origin()) *
		tan(_camera.x_fov());
	real_t tx = -l * dx / _viewport.width();
	real_t ty = l * dy / _viewport.height();
	_camera.translate((tx * _camera.right()) + (ty * _camera.up()));
}


void View::scale_world(const Point3 &origin, int dx, int dy)
{
	(void) dx;
	if (dy == 0) {
		return;
	}
	real_t t = 2.0 * (real_t) dy / (real_t) _viewport.width();
	_camera.translate(t * (origin - _camera.origin()));
}


void View::scale_world(real_t factor)
{
	_camera.translate(factor * _camera.towards());
}


void View::load_camera(const vector<Light> &headlights) const
{
	_camera.load(_viewport, headlights);
}


void View::load_camera() const
{
	_camera.load(_viewport);
}


Point3 View::viewport_point(int x, int y) const
{
	Point2 vc = _viewport.center();
	real_t dx = (real_t) (x - vc.x()) / (real_t) (_viewport.width() / 2);
	real_t dy = (real_t) (y - vc.y()) / (real_t) (_viewport.height() / 2);
	Point3 wc = (_camera.origin() +
		     _camera.near_distance() * _camera.towards());
	Vector3 r = (_camera.near_distance() * tan(_camera.x_fov()) *
		     _camera.right());
	Vector3 u = (_camera.near_distance() * tan(_camera.y_fov()) *
		     _camera.up());
	return wc + (dx * r) + (dy * u);
}


Ray3 View::world_ray(int x, int y) const
{
	return Ray3(_camera.origin(), viewport_point(x, y));
}


Point2 View::viewport_point(const Point3 &p) const
{
	// equivalent to gluProject
	Point3 q = _camera.world_to_camera(p);
//	assert(!real::is_zero(q.z()));
	real_t dx = q.x() / (-q.z() * tan(_camera.x_fov()));
	real_t dy = q.y() / (-q.z() * tan(_camera.y_fov()));
	real_t x = _viewport.x_center() + dx * _viewport.width() * 0.5;
	real_t y = _viewport.y_center() + dy * _viewport.height() * 0.5;
	return Point2(x, y);
}


void View::set_viewport(const Viewport &vp)
{
    _viewport = vp;
    real_t fovx = _camera.x_fov();
    real_t w = vp.width();
    real_t h = vp.height();
    real_t fovy = atan((double)h/(double)w*tan(fovx));
    _camera.reset_fov(fovx, fovy);
}


END_GTBMOUSENS
